@@ -14,6 +14,7 @@
 //
 // $URL$
 // $Id$
+// SPDX-License-Identifier: GPL-3.0+
 // 
 //
 // Author(s)     : Jane Tournois
@@ -22,6 +23,9 @@
 #ifndef CGAL_POLYGON_MESH_PROCESSING_COMPUTE_NORMAL_H
 #define CGAL_POLYGON_MESH_PROCESSING_COMPUTE_NORMAL_H
 
+#include <CGAL/license/Polygon_mesh_processing/Compute_normal.h>
+
+
 #include <CGAL/boost/graph/helpers.h>
 #include <CGAL/boost/graph/properties.h>
 #include <boost/graph/graph_traits.hpp>
@@ -40,19 +44,38 @@ namespace Polygon_mesh_processing{
 
 namespace internal {
 
-  template<typename Point>
-  typename CGAL::Kernel_traits<Point>::Kernel::Vector_3
-  triangle_normal(const Point& p0, const Point& p1, const Point& p2)
+  template <class GT>
+  void normalize(typename GT::Vector_3& v, const GT& traits)
   {
-    return CGAL::cross_product(p2 - p1, p0 - p1);
+    typename GT::FT norm = CGAL::approximate_sqrt(
+        traits.compute_squared_length_3_object()(v));
+    v = traits.construct_divided_vector_3_object()(v, norm);
+  }
+
+  template<typename Point
+         , typename GT>
+  typename GT::Vector_3
+  triangle_normal(const Point& p0, const Point& p1, const Point& p2
+                , const GT& traits)
+  {
+    typename GT::Vector_3 n = traits.construct_cross_product_vector_3_object()(
+      traits.construct_vector_3_object()(p1, p2),
+      traits.construct_vector_3_object()(p1, p0));
+
+    //cross-product(AB, AC)'s norm is the area of the parallelogram 
+    //formed by these 2 vectors.
+    //the triangle's area is half of it
+    return traits.construct_scaled_vector_3_object()(n, 0.5);
   }
 }
 
-template<typename Point, typename PM, typename VertexPointMap, typename Vector>
+template<typename Point, typename PM, typename VertexPointMap, typename Vector
+       , typename GT>
 void sum_normals(const PM& pmesh,
                  typename boost::graph_traits<PM>::face_descriptor f,
                  VertexPointMap vpmap,
-                 Vector& sum)
+                 Vector& sum,
+                 const GT& traits)
 {
   typedef typename boost::graph_traits<PM>::vertex_descriptor   vertex_descriptor;
   typedef typename boost::graph_traits<PM>::halfedge_descriptor halfedge_descriptor;
@@ -65,8 +88,8 @@ void sum_normals(const PM& pmesh,
     const Point& pvn  = get(vpmap, target(he, pmesh));
     const Point& pvnn = get(vpmap, target(next(he, pmesh), pmesh));
 
-    Vector n = internal::triangle_normal(pv, pvn, pvnn);
-    sum = sum + n;
+    Vector n = internal::triangle_normal(pv, pvn, pvnn, traits);
+    sum = traits.construct_sum_of_vectors_3_object()(sum, n);
 
     he = next(he, pmesh);
   }
@@ -76,23 +99,28 @@ void sum_normals(const PM& pmesh,
 /**
 * \ingroup PMP_normal_grp
 * computes the outward unit vector normal to face `f`.
-* @tparam PolygonMesh a model of `FaceGraph` that has an internal property map
-*         for `CGAL::vertex_point_t`
-* @tparam NamedParameters a sequence of \ref namedparameters
+* @tparam PolygonMesh a model of `FaceGraph`
+* @tparam NamedParameters a sequence of \ref pmp_namedparameters "Named Parameters"
 *
 * @param f the face on which the normal is computed
 * @param pmesh the polygon mesh containing `f`
-* @param np optional sequence of \ref namedparameters among the ones listed below
+* @param np optional sequence of \ref pmp_namedparameters "Named Parameters" among the ones listed below
 *
 * \cgalNamedParamsBegin
-*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh` \cgalParamEnd
-*    \cgalParamBegin{geom_traits} a geometric traits class instance \cgalParamEnd
+*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh`.
+*   If this parameter is omitted, an internal property map for
+*   `CGAL::vertex_point_t` must be available in `PolygonMesh`\cgalParamEnd
+*    \cgalParamBegin{geom_traits} an instance of a geometric traits class, model of `Kernel`\cgalParamEnd
 * \cgalNamedParamsEnd
 *
 * @return the computed normal. The return type is a 3D vector type. It is
-* either deduced from the `geom_traits` \ref namedparameters if provided,
+* either deduced from the `geom_traits` \ref pmp_namedparameters "Named Parameters" if provided,
 * or from the geometric traits class deduced from the point property map
 * of `pmesh`.
+*
+* \warning This function involves a square root computation.
+* If `Kernel::FT` does not have a `sqrt()` operation, the square root computation
+* will be done approximately.
 */
 template <typename PolygonMesh, typename NamedParameters>
 #ifdef DOXYGEN_RUNNING
@@ -104,43 +132,49 @@ compute_face_normal(typename boost::graph_traits<PolygonMesh>::face_descriptor f
                     , const PolygonMesh& pmesh
                     , const NamedParameters& np)
 {
-  typedef typename GetGeomTraits<PolygonMesh, NamedParameters>::type Kernel;
-  typedef typename Kernel::FT FT;
-  typedef typename Kernel::Point_3 Point;
-  typedef typename Kernel::Vector_3 Vector;
+  typedef typename GetGeomTraits<PolygonMesh, NamedParameters>::type GT;
+  typedef typename GT::Point_3 Point;
+  typedef typename GT::Vector_3 Vector;
 
+  using boost::choose_param;
   using boost::get_param;
-  using boost::choose_const_pmap;
 
-  Vector normal = CGAL::NULL_VECTOR;
+  GT traits = choose_param(get_param(np, internal_np::geom_traits), GT());
+
+  Vector normal = traits.construct_vector_3_object()(CGAL::NULL_VECTOR);
   sum_normals<Point>(pmesh, f
-    , choose_const_pmap(get_param(np, CGAL::vertex_point), pmesh, CGAL::vertex_point)
-    , normal);
+    , choose_param(get_param(np, internal_np::vertex_point), get_const_property_map(CGAL::vertex_point, pmesh))
+    , normal
+    , traits);
 
-  if (normal == CGAL::NULL_VECTOR)
-    return normal;
-  else
-  return normal / FT( std::sqrt( to_double(normal * normal) ) );
+  if (!typename GT::Equal_3()(normal, CGAL::NULL_VECTOR))
+    internal::normalize(normal, traits);
+
+  return normal;
 }
 
 /**
 * \ingroup PMP_normal_grp
 * computes the outward unit vector normal for all faces of the polygon mesh.
-* @tparam PolygonMesh a model of `FaceGraph` that has an internal property map
-*         for `CGAL::vertex_point_t`
+* @tparam PolygonMesh a model of `FaceGraph`
 * @tparam FaceNormalMap a model of `WritablePropertyMap` with
     `boost::graph_traits<PolygonMesh>::%face_descriptor` as key type and
     `Kernel::Vector_3` as value type.
 *
 * @param pmesh the polygon mesh
 * @param fnm the property map in which the normals are written
-* @param np optional sequence of \ref namedparameters among the ones listed below
+* @param np optional sequence of \ref pmp_namedparameters "Named Parameters" among the ones listed below
 *
 * \cgalNamedParamsBegin
-*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh` \cgalParamEnd
-*    \cgalParamBegin{geom_traits} a geometric traits class instance \cgalParamEnd
+*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh`.
+*   If this parameter is omitted, an internal property map for
+*   `CGAL::vertex_point_t` must be available in `PolygonMesh`\cgalParamEnd
+*    \cgalParamBegin{geom_traits} an instance of a geometric traits class, model of `Kernel`\cgalParamEnd
 * \cgalNamedParamsEnd
 *
+* \warning This function involves a square root computation.
+* If `Kernel::FT` does not have a `sqrt()` operation, the square root computation
+* will be done approximately.
 */
 template <typename PolygonMesh
           , typename FaceNormalMap
@@ -162,22 +196,27 @@ compute_face_normals(const PolygonMesh& pmesh
 /**
 * \ingroup PMP_normal_grp
 * computes the unit normal at vertex `v` as the average of the normals of incident faces.
-* @tparam PolygonMesh a model of `FaceGraph` that has an internal property map
-*         for `CGAL::vertex_point_t`
+* @tparam PolygonMesh a model of `FaceGraph`
 *
 * @param v the vertex at which the normal is computed
 * @param pmesh the polygon mesh containing `v`
-* @param np optional sequence of \ref namedparameters among the ones listed below
+* @param np optional sequence of \ref pmp_namedparameters "Named Parameters" among the ones listed below
 *
 * \cgalNamedParamsBegin
-*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh` \cgalParamEnd
-*    \cgalParamBegin{geom_traits} a geometric traits class instance \cgalParamEnd
+*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh`.
+*   If this parameter is omitted, an internal property map for
+*   `CGAL::vertex_point_t` must be available in `PolygonMesh`\cgalParamEnd
+*    \cgalParamBegin{geom_traits} an instance of a geometric traits class, model of `Kernel`\cgalParamEnd
 * \cgalNamedParamsEnd
 *
 * @return the computed normal. The return type is a 3D vector type. It is
-* either deduced from the `geom_traits` \ref namedparameters if provided,
+* either deduced from the `geom_traits` \ref pmp_namedparameters "Named Parameters" if provided,
 * or the geometric traits class deduced from the point property map
 * of `pmesh`.
+*
+* \warning This function involves a square root computation.
+* If `Kernel::FT` does not have a `sqrt()` operation, the square root computation
+* will be done approximately.
 */
 template<typename PolygonMesh, typename NamedParameters>
 #ifdef DOXYGEN_RUNNING
@@ -190,67 +229,69 @@ compute_vertex_normal(typename boost::graph_traits<PolygonMesh>::vertex_descript
                       const NamedParameters& np
                       )
 {
-  typedef typename GetGeomTraits<PolygonMesh, NamedParameters>::type Kernel;
-  typedef typename Kernel::FT FT;
+  using boost::choose_param;
+
+  typedef typename GetGeomTraits<PolygonMesh, NamedParameters>::type GT;
+  typedef typename GT::Vector_3 Vector;
+  GT traits = choose_param(get_param(np, internal_np::geom_traits), GT());
 
   typedef typename GetFaceNormalMap<PolygonMesh, NamedParameters>::NoMap DefaultMap;
-
   typedef typename boost::lookup_named_param_def <
-    CGAL::face_normal_t,
+    internal_np::face_normal_t,
     NamedParameters,
     DefaultMap> ::type FaceNormalMap;
-
-  FaceNormalMap fnmap
-    = boost::choose_param(get_param(np, face_normal), DefaultMap());
-
+  FaceNormalMap fnmap = choose_param(get_param(np, internal_np::face_normal), DefaultMap());
   bool fnmap_valid
     = !boost::is_same<FaceNormalMap,
                       DefaultMap
                      >::value;
 
-  typedef typename Kernel::Vector_3 Vector;
   typedef typename boost::graph_traits<PolygonMesh>::halfedge_descriptor halfedge_descriptor;
 
-  Vector normal = CGAL::NULL_VECTOR;
+  Vector normal = traits.construct_vector_3_object()(CGAL::NULL_VECTOR);
+
   halfedge_descriptor he = halfedge(v, pmesh);
   // handle isolated vertices
   if (he==boost::graph_traits<PolygonMesh>::null_halfedge()) return normal;
   halfedge_descriptor end = he;
   do
-    {
+  {
     if (!is_border(he, pmesh))
     {
       Vector n = fnmap_valid ? get(fnmap, face(he, pmesh))
                              : compute_face_normal(face(he, pmesh), pmesh, np);
-      normal = normal + n;
+      normal = traits.construct_sum_of_vectors_3_object()(normal, n);
     }
     he = opposite(next(he, pmesh), pmesh);
   } while (he != end);
 
-  if (normal == CGAL::NULL_VECTOR)
-    return normal;
-  else
-  return normal / FT( std::sqrt( to_double(normal * normal)  ) );
+  if ( ! typename GT::Equal_3()(normal, CGAL::NULL_VECTOR))
+    internal::normalize(normal, traits);
+  return normal;
 }
 
 /**
 * \ingroup PMP_normal_grp
 * computes the outward unit vector normal for all vertices of the polygon mesh.
-* @tparam PolygonMesh a model of `FaceListGraph` that has an internal property map
-*         for `CGAL::vertex_point_t`
+* @tparam PolygonMesh a model of `FaceListGraph`
 * @tparam VertexNormalMap a model of `WritablePropertyMap` with
     `boost::graph_traits<PolygonMesh>::%vertex_descriptor` as key type and
     the return type of `compute_vertex_normal()` as value type.
 *
 * @param pmesh the polygon mesh
 * @param vnm the property map in which the normals are written
-* @param np optional sequence of \ref namedparameters among the ones listed below
+* @param np optional sequence of \ref pmp_namedparameters "Named Parameters" among the ones listed below
 *
 * \cgalNamedParamsBegin
-*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh` \cgalParamEnd
-*    \cgalParamBegin{geom_traits} a geometric traits class instance \cgalParamEnd
+*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh`.
+*   If this parameter is omitted, an internal property map for
+*   `CGAL::vertex_point_t` must be available in `PolygonMesh`\cgalParamEnd
+*    \cgalParamBegin{geom_traits} an instance of a geometric traits class, model of `Kernel`\cgalParamEnd
 * \cgalNamedParamsEnd
 *
+* \warning This function involves a square root computation.
+* If `Kernel::FT` does not have a `sqrt()` operation, the square root computation
+* will be done approximately.
 */
 template <typename PolygonMesh
           , typename VertexNormalMap
@@ -274,8 +315,7 @@ compute_vertex_normals(const PolygonMesh& pmesh
 /**
 * \ingroup PMP_normal_grp
 * computes the outward unit vector normal for all vertices and faces of the polygon mesh.
-* @tparam PolygonMesh a model of `FaceListGraph` that has an internal property map
-*         for `CGAL::vertex_point_t`
+* @tparam PolygonMesh a model of `FaceListGraph`
 * @tparam VertexNormalMap a model of `WritablePropertyMap` with
     `boost::graph_traits<PolygonMesh>::%vertex_descriptor` as key type and
     `Kernel::Vector_3` as value type.
@@ -286,13 +326,18 @@ compute_vertex_normals(const PolygonMesh& pmesh
 * @param pmesh the polygon mesh
 * @param vnm the property map in which the vertex normals are written
 * @param fnm the property map in which the face normals are written
-* @param np optional sequence of \ref namedparameters among the ones listed below
+* @param np optional sequence of \ref pmp_namedparameters "Named Parameters" among the ones listed below
 *
 * \cgalNamedParamsBegin
-*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh` \cgalParamEnd
-*    \cgalParamBegin{geom_traits} a geometric traits class instance \cgalParamEnd
+*    \cgalParamBegin{vertex_point_map} the property map with the points associated to the vertices of `pmesh`.
+*   If this parameter is omitted, an internal property map for
+*   `CGAL::vertex_point_t` must be available in `PolygonMesh`\cgalParamEnd
+*    \cgalParamBegin{geom_traits} an instance of a geometric traits class, model of `Kernel`\cgalParamEnd
 * \cgalNamedParamsEnd
 *
+* \warning This function involves a square root computation.
+* If `Kernel::FT` does not have a `sqrt()` operation, the square root computation
+* will be done approximately.
 */
 template <typename PolygonMesh
           , typename VertexNormalMap
